@@ -21,7 +21,7 @@ def contract_path(*operands, **kwargs):
         Specifies the subscripts for summation.
     *operands : list of array_like
         These are the arrays for the operation.
-    path : bool or list, optional (default: ``greedy``)
+    path : bool or list, optional (default: ``auto``)
         Choose the type of path.
 
         - if a list is given uses this as the path.
@@ -31,6 +31,8 @@ def contract_path(*operands, **kwargs):
         - 'optimal' An algorithm that tries all possible ways of
           contracting the listed tensors. Scales exponentially with
           the number of terms in the contraction.
+        - 'auto' Use the optimal approach for small numbers of terms (currently
+          4 or less), else use the greedy approach.
 
     use_blas : bool
         Use BLAS functions or not
@@ -110,7 +112,7 @@ def contract_path(*operands, **kwargs):
     if len(unknown_kwargs):
         raise TypeError("einsum_path: Did not understand the following kwargs: %s" % unknown_kwargs)
 
-    path_type = kwargs.pop('path', 'greedy')
+    path_type = kwargs.pop('path', 'auto')
     memory_limit = kwargs.pop('memory_limit', None)
 
     # Hidden option, only einsum should call this
@@ -123,14 +125,14 @@ def contract_path(*operands, **kwargs):
     # Build a few useful list and sets
     input_list = input_subscripts.split(',')
     input_sets = [set(x) for x in input_list]
+    input_shps = [x.shape for x in operands]
     output_set = set(output_subscript)
     indices = set(input_subscripts.replace(',', ''))
 
     # Get length of each unique dimension and ensure all dimensions are correct
     dimension_dict = {}
-    broadcast_indices = [[] for x in range(len(input_list))]
     for tnum, term in enumerate(input_list):
-        sh = operands[tnum].shape
+        sh = input_shps[tnum]
 
         if len(sh) != len(term):
             raise ValueError("Einstein sum subscript %s does not contain the "
@@ -138,11 +140,7 @@ def contract_path(*operands, **kwargs):
         for cnum, char in enumerate(term):
             dim = sh[cnum]
 
-            # Build out broadcast indices
-            if dim == 1:
-                broadcast_indices[tnum].append(char)
-
-            if char in dimension_dict.keys():
+            if char in dimension_dict:
                 # For broadcasting cases we always want the largest dim size
                 if dimension_dict[char] == 1:
                     dimension_dict[char] = dim
@@ -152,13 +150,8 @@ def contract_path(*operands, **kwargs):
             else:
                 dimension_dict[char] = dim
 
-    # Convert broadcast inds to sets
-    broadcast_indices = [set(x) for x in broadcast_indices]
-
     # Compute size of each input array plus the output array
-    size_list = []
-    for term in input_list + [output_subscript]:
-        size_list.append(helpers.compute_size_by_dict(term, dimension_dict))
+    size_list = [helpers.compute_size_by_dict(term, dimension_dict) for term in input_list + [output_subscript]]
     out_size = max(size_list)
 
     if memory_limit is None:
@@ -172,29 +165,31 @@ def contract_path(*operands, **kwargs):
         else:
             memory_arg = int(memory_limit)
 
+    num_ops = len(input_list)
+
     # Compute naive cost
     # This isnt quite right, need to look into exactly how einsum does this
     # indices_in_input = input_subscripts.replace(',', '')
-    # inne
+
     inner_product = (sum(len(x) for x in input_sets) - len(indices)) > 0
-    naive_cost = helpers.flop_count(indices, inner_product, len(input_list), dimension_dict)
+    naive_cost = helpers.flop_count(indices, inner_product, num_ops, dimension_dict)
 
     # Compute the path
     if not isinstance(path_type, str):
         path = path_type
-    elif len(input_list) == 1:
+    elif num_ops == 1:
         # Nothing to be optimized
         path = [(0, )]
-    elif len(input_list) == 2:
+    elif num_ops == 2:
         # Nothing to be optimized
         path = [(0, 1)]
     elif indices == output_set:
         # If no rank reduction leave it to einsum
-        path = [tuple(range(len(input_list)))]
-    elif path_type in ["greedy", "opportunistic"]:
-        path = paths.greedy(input_sets, output_set, dimension_dict, memory_arg)
-    elif path_type == "optimal":
+        path = [tuple(range(num_ops))]
+    elif path_type == "optimal" or (path_type == "auto" and num_ops <= 4):
         path = paths.optimal(input_sets, output_set, dimension_dict, memory_arg)
+    elif path_type in ("greedy", "opportunistic", "auto"):
+        path = paths.greedy(input_sets, output_set, dimension_dict, memory_arg)
     else:
         raise KeyError("Path name %s not found" % path_type)
 
@@ -217,17 +212,11 @@ def contract_path(*operands, **kwargs):
         scale_list.append(len(idx_contract))
         size_list.append(helpers.compute_size_by_dict(out_inds, dimension_dict))
 
-        bcast = set()
-        tmp_inputs = []
-        for x in contract_inds:
-            tmp_inputs.append(input_list.pop(x))
-            bcast |= broadcast_indices.pop(x)
+        tmp_inputs = [input_list.pop(x) for x in contract_inds]
+        tmp_shapes = [input_shps.pop(x) for x in contract_inds]
 
-        new_bcast_inds = bcast - idx_removed
-
-        # If were broadcasting, nix blas
-        if use_blas and not len(idx_removed & bcast):
-            do_blas = blas.can_blas(tmp_inputs, out_inds, idx_removed)
+        if use_blas:
+            do_blas = blas.can_blas(tmp_inputs, out_inds, idx_removed, tmp_shapes)
         else:
             do_blas = False
 
@@ -235,11 +224,15 @@ def contract_path(*operands, **kwargs):
         if (cnum - len(path)) == -1:
             idx_result = output_subscript
         else:
-            sort_result = [(dimension_dict[ind], ind) for ind in out_inds]
-            idx_result = "".join([x[1] for x in sorted(sort_result)])
+            # use tensordot order to minimize transpositions
+            all_input_inds = "".join(tmp_inputs)
+            idx_result = "".join(sorted(out_inds, key=all_input_inds.find))
+
+        shp_result = parser.find_output_shape(tmp_inputs, tmp_shapes, idx_result)
 
         input_list.append(idx_result)
-        broadcast_indices.append(new_bcast_inds)
+        input_shps.append(shp_result)
+
         einsum_str = ",".join(tmp_inputs) + "->" + idx_result
 
         contraction = (contract_inds, idx_removed, einsum_str, input_list[:], do_blas)
@@ -301,7 +294,7 @@ def _transpose(x, axes, backend='numpy'):
     """
     try:
         return x.transpose(axes)
-    except AttributeError:
+    except (AttributeError, TypeError):
         # some libraries don't implement method version
         fn = backends.get_func('transpose', backend)
         return fn(x, axes)
@@ -402,7 +395,8 @@ def contract(*operands, **kwargs):
     use_blas = kwargs.pop('use_blas', True)
     memory_limit = kwargs.pop('memory_limit', None)
     backend = kwargs.pop('backend', 'numpy')
-    gen_expression = kwargs.pop('gen_expression', False)
+    gen_expression = kwargs.pop('_gen_expression', False)
+    constants_dict = kwargs.pop('_constants_dict', {})
 
     # Make sure remaining keywords are valid for einsum
     unknown_kwargs = [k for (k, v) in kwargs.items() if k not in valid_einsum_kwargs]
@@ -418,17 +412,17 @@ def contract(*operands, **kwargs):
 
     # check if performing contraction or just building expression
     if gen_expression:
-        return ContractExpression(full_str, contraction_list, **einsum_kwargs)
+        return ContractExpression(full_str, contraction_list, constants_dict, **einsum_kwargs)
 
     return _core_contract(operands, contraction_list, backend=backend, **einsum_kwargs)
 
 
-def _core_contract(operands, contraction_list, backend='numpy', **einsum_kwargs):
+def _core_contract(operands, contraction_list, backend='numpy', evaluate_constants=False, **einsum_kwargs):
     """Inner loop used to perform an actual contraction given the output
     from a ``contract_path(..., einsum_call=True)`` call.
     """
 
-    # Special handeling if out is specified
+    # Special handling if out is specified
     out_array = einsum_kwargs.pop('out', None)
     specified_out = out_array is not None
 
@@ -438,9 +432,13 @@ def _core_contract(operands, contraction_list, backend='numpy', **einsum_kwargs)
     # Start contraction loop
     for num, contraction in enumerate(contraction_list):
         inds, idx_rm, einsum_str, remaining, blas_flag = contraction
-        tmp_operands = []
-        for x in inds:
-            tmp_operands.append(operands.pop(x))
+
+        # check if we are performing the pre-pass of an expression with constants,
+        #     if so, break out upon finding first non-constant (None) operand
+        if evaluate_constants and any(operands[x] is None for x in inds):
+            return operands, contraction_list[num:]
+
+        tmp_operands = [operands.pop(x) for x in inds]
 
         # Do we need to deal with the output?
         handle_out = specified_out and ((num + 1) == len(contraction_list))
@@ -452,9 +450,7 @@ def _core_contract(operands, contraction_list, backend='numpy', **einsum_kwargs)
             input_str, results_index = einsum_str.split('->')
             input_left, input_right = input_str.split(',')
 
-            tensor_result = input_left + input_right
-            for s in idx_rm:
-                tensor_result = tensor_result.replace(s, "")
+            tensor_result = "".join(s for s in input_left + input_right if s not in idx_rm)
 
             # Find indices to contract over
             left_pos, right_pos = [], []
@@ -483,7 +479,7 @@ def _core_contract(operands, contraction_list, backend='numpy', **einsum_kwargs)
             # Do the contraction
             new_view = _einsum(einsum_str, *tmp_operands, backend=backend, **einsum_kwargs)
 
-        # Append new items and derefernce what we can
+        # Append new items and dereference what we can
         operands.append(new_view)
         del tmp_operands, new_view
 
@@ -493,35 +489,109 @@ def _core_contract(operands, contraction_list, backend='numpy', **einsum_kwargs)
         return operands[0]
 
 
+def format_const_einsum_str(einsum_str, constants):
+    """Add brackets to the constant terms in ``einsum_str``. For example:
+
+        >>> format_const_einsum_str('ab,bc,cd->ad', [0, 2])
+        'bc,[ab,cd]->ad'
+
+    No-op if there are no constants.
+    """
+    if not constants:
+        return einsum_str
+
+    if "->" in einsum_str:
+        lhs, rhs = einsum_str.split('->')
+        arrow = "->"
+    else:
+        lhs, rhs, arrow = einsum_str, "", ""
+
+    wrapped_terms = ["[{}]".format(t) if i in constants else t for i, t in enumerate(lhs.split(','))]
+
+    formatted_einsum_str = "{}{}{}".format(','.join(wrapped_terms), arrow, rhs)
+
+    # merge adjacent constants
+    formatted_einsum_str = formatted_einsum_str.replace("],[", ',')
+    return formatted_einsum_str
+
+
 class ContractExpression:
     """Helper class for storing an explicit ``contraction_list`` which can
     then be repeatedly called solely with the array arguments.
     """
 
-    def __init__(self, contraction, contraction_list, **einsum_kwargs):
-        self.contraction = contraction
+    def __init__(self, contraction, contraction_list, constants_dict, **einsum_kwargs):
         self.contraction_list = contraction_list
         self.einsum_kwargs = einsum_kwargs
-        self.num_args = len(contraction.split('->')[0].split(','))
+        self.contraction = format_const_einsum_str(contraction, constants_dict.keys())
 
-    def _normal_contract(self, arrays, out=None, backend='numpy'):
+        # need to know _full_num_args to parse constants with, and num_args to call with
+        self._full_num_args = contraction.count(',') + 1
+        self.num_args = self._full_num_args - len(constants_dict)
+
+        # likewise need to know full contraction list
+        self._full_contraction_list = contraction_list
+
+        self._constants_dict = constants_dict
+        self._evaluated_constants = {}
+        self._backend_expressions = {}
+
+    def evaluate_constants(self, backend='numpy'):
+        """Convert any constant operands to the correct backend form, and
+        perform as many contractions as possible to create a new list of
+        operands, stored in ``self._evaluated_constants[backend]``. This also
+        makes sure ``self.contraction_list`` only contains the remaining,
+        non-const operations.
+        """
+        # prepare a list of operands, with `None` for non-consts
+        tmp_const_ops = [self._constants_dict.get(i, None) for i in range(self._full_num_args)]
+
+        # get the new list of operands with constant operations performed, and remaining contractions
+        new_ops, new_contraction_list = self(*tmp_const_ops, backend=backend, evaluate_constants=True)
+        self._evaluated_constants[backend] = new_ops
+        self.contraction_list = new_contraction_list
+
+    def _get_evaluated_constants(self, backend):
+        """Retrieve or generate the cached list of constant operators (mixed
+        in with None representing non-consts) and the remaining contraction
+        list.
+        """
+        try:
+            return self._evaluated_constants[backend]
+        except KeyError:
+            self.evaluate_constants(backend)
+            return self._evaluated_constants[backend]
+
+    def _get_backend_expression(self, arrays, backend):
+        try:
+            return self._backend_expressions[backend]
+        except KeyError:
+            fn = backends.build_expression(backend, arrays, self)
+            self._backend_expressions[backend] = fn
+            return fn
+
+    def _contract(self, arrays, out=None, backend='numpy', evaluate_constants=False):
         """The normal, core contraction.
         """
-        return _core_contract(list(arrays), self.contraction_list, out=out,
-                              backend=backend, **self.einsum_kwargs)
+        contraction_list = self._full_contraction_list if evaluate_constants else self.contraction_list
 
-    def _convert_contract(self, arrays, out, backend):
+        return _core_contract(list(arrays), contraction_list, out=out, backend=backend,
+                              evaluate_constants=evaluate_constants, **self.einsum_kwargs)
+
+    def _contract_with_conversion(self, arrays, out, backend, evaluate_constants=False):
         """Special contraction, i.e. contraction with a different backend
-        but converting to and from that backend. Checks for
-        ``self._{backend}_contract``, generates it if is missing, then calls it
+        but converting to and from that backend. Retrieves or generates a
+        cached expression using ``arrays`` as templates, then calls it
         with ``arrays``.
+
+        If ``evaluate_constants=True``, perform a partial contraction that
+        prepares the constant tensors and operations with the right backend.
         """
-        convert_fn = "_{}_contract".format(backend)
+        # convert consts to correct type & find reduced contraction list
+        if evaluate_constants:
+            return backends.evaluate_constants(backend, arrays, self)
 
-        if not hasattr(self, convert_fn):
-            setattr(self, convert_fn, backends.build_expression(backend, arrays, self))
-
-        result = getattr(self, convert_fn)(*arrays)
+        result = self._get_backend_expression(arrays, backend)(*arrays)
 
         if out is not None:
             out[()] = result
@@ -543,24 +613,34 @@ class ContractExpression:
             are supplied then try to convert them to and from the correct
             backend array type.
         """
-
-        if len(arrays) != self.num_args:
-            raise ValueError("This `ContractExpression` takes exactly %s array arguments "
-                             "but received %s." % (self.num_args, len(arrays)))
-
-        backend = kwargs.pop('backend', 'numpy')
         out = kwargs.pop('out', None)
+        backend = kwargs.pop('backend', 'numpy')
+        evaluate_constants = kwargs.pop('evaluate_constants', False)
+
         if kwargs:
             raise ValueError("The only valid keyword arguments to a `ContractExpression` "
                              "call are `out=` or `backend=`. Got: %s." % kwargs)
 
+        correct_num_args = self._full_num_args if evaluate_constants else self.num_args
+
+        if len(arrays) != correct_num_args:
+            raise ValueError("This `ContractExpression` takes exactly %s array arguments "
+                             "but received %s." % (self.num_args, len(arrays)))
+
+        if self._constants_dict and not evaluate_constants:
+            # fill in the missing non-constant terms with newly supplied arrays
+            ops_var, ops_const = iter(arrays), self._get_evaluated_constants(backend)
+            ops = [next(ops_var) if op is None else op for op in ops_const]
+        else:
+            ops = arrays
+
         try:
             # Check if the backend requires special preparation / calling
             #   but also ignore non-numpy arrays -> assume user wants same type back
-            if backend in backends.CONVERT_BACKENDS and isinstance(arrays[0], np.ndarray):
-                return self._convert_contract(arrays, out, backend)
+            if backend in backends.CONVERT_BACKENDS and any(isinstance(x, np.ndarray) for x in arrays):
+                return self._contract_with_conversion(ops, out, backend, evaluate_constants=evaluate_constants)
 
-            return self._normal_contract(arrays, out, backend)
+            return self._contract(ops, out, backend, evaluate_constants=evaluate_constants)
 
         except ValueError as err:
             original_msg = str(err.args) if err.args else ""
@@ -571,10 +651,14 @@ class ContractExpression:
             raise
 
     def __repr__(self):
-        return "ContractExpression('%s')" % self.contraction
+        if self._constants_dict:
+            constants_repr = ", constants={}".format(sorted(self._constants_dict))
+        else:
+            constants_repr = ""
+        return "<ContractExpression('{}'{})>".format(self.contraction, constants_repr)
 
     def __str__(self):
-        s = "<ContractExpression> for '%s':" % self.contraction
+        s = self.__repr__()
         for i, c in enumerate(self.contraction_list):
             s += "\n  %i.  " % (i + 1)
             s += "'%s'" % c[2] + (" [%s]" % c[-1] if c[-1] else "")
@@ -583,13 +667,11 @@ class ContractExpression:
         return s
 
 
-class _ShapeOnly(np.ndarray):
+def shape_only(shape):
     """Dummy ``numpy.ndarray`` which has a shape only - for generating
     contract expressions.
     """
-
-    def __init__(self, shape):
-        self.shape = shape
+    return np.broadcast_to(np.nan, shape)
 
 
 def contract_expression(subscripts, *shapes, **kwargs):
@@ -602,6 +684,13 @@ def contract_expression(subscripts, *shapes, **kwargs):
         Specifies the subscripts for summation.
     shapes : sequence of integer tuples
         Shapes of the arrays to optimize the contraction for.
+    constants : sequence of int, optional
+        The indices of any constant arguments in ``shapes``, in which case the
+        actual array should be supplied at that position rather than just a
+        shape. If these are specified, then constant parts of the contraction
+        between calls will be reused. Additionally, if a gpu-enabled backend is
+        used for example, then the constant tensors will be kept on the gpu,
+        minimizing transfers.
     kwargs :
         Passed on to ``contract_path`` or ``einsum``. See ``contract``.
 
@@ -621,15 +710,29 @@ def contract_expression(subscripts, *shapes, **kwargs):
     - The generated expression will work with any arrays which have
       the same rank (number of dimensions) as the original shapes, however, if
       the actual sizes are different, the expression may no longer be optimal.
+    - Constant operations will be computed upon first call with a particular
+      backend, then subsequently reused.
 
     Examples
     --------
 
-    >>> expr = contract_expression("ab,bc->ac", (3, 4), (4, 5))
-    >>> a, b = np.random.rand(3, 4), np.random.rand(4, 5)
-    >>> c = expr(a, b)
-    >>> np.allclose(c, a @ b)
-    True
+    Basic usage:
+
+        >>> expr = contract_expression("ab,bc->ac", (3, 4), (4, 5))
+        >>> a, b = np.random.rand(3, 4), np.random.rand(4, 5)
+        >>> c = expr(a, b)
+        >>> np.allclose(c, a @ b)
+        True
+
+    Supply ``a`` as a constant:
+
+        >>> expr = contract_expression("ab,bc->ac", a, (4, 5), constants=[0])
+        >>> expr
+        <ContractExpression('[ab],bc->ac', constants=[0])>
+
+        >>> c = expr(b)
+        >>> np.allclose(c, a @ b)
+        True
 
     """
     if not kwargs.get('optimize', True):
@@ -640,6 +743,14 @@ def contract_expression(subscripts, *shapes, **kwargs):
             raise ValueError("'%s' should only be specified when calling a "
                              "`ContractExpression`, not when building it." % arg)
 
-    dummy_arrays = [_ShapeOnly(s) for s in shapes]
+    kwargs['_gen_expression'] = True
 
-    return contract(subscripts, *dummy_arrays, gen_expression=True, **kwargs)
+    # build dict of constant indices mapped to arrays
+    constants = kwargs.pop('constants', ())
+    constants_dict = {i: shapes[i] for i in constants}
+    kwargs['_constants_dict'] = constants_dict
+
+    # apart from constant arguments, make dummy arrays
+    dummy_arrays = [s if i in constants else shape_only(s) for i, s in enumerate(shapes)]
+
+    return contract(subscripts, *dummy_arrays, **kwargs)
